/*
 * main.c
 *
 *  Created on: Oct 5, 2015
 *	  Author: boris
 */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

#ifndef F_CPU
#define F_CPU 8000000UL
#endif

#define freq_0 555

#define sevens_underscore	0b0001000
#define sevens_H			0b0110111

// table for converting binary to 7 segment code (DP @ MSB)
uint8_t sedmisegment[] = {
  0b01111110,	// 0
  0b00110000,	// 1
  0b01101101,	// 2
  0b01111001,	// 3
  0b00110011,	// 4
  0b01011011,	// 5
  0b01011111,	// 6
  0b01110010,	// 7
  0b01111111,	// 8
  0b01111011,	// 9
  0b01111101,	// 10/A
  0b00011111,	// 11/B
  0b00001101,	// 12/C
  0b00111101,	// 13/D
  0b01101111,	// 14/E
  0b01000111,	// 15/F
  0b10000000	// 16/ERROR
};

void spi_send(uint8_t data){
	SPDR = data;
	// if SPDR write collision:
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE){
		if(SPSR & (1<<WCOL)){
			while(!(SPSR & (1<<SPIF)));		// wait for transmission to finish
			SPDR = data;					// try again
		}
		while(!(SPSR & (1<<SPIF)));
	}
}

// send out tuning word to AD9833 over SPI
// freq in tenths of Hz:
void AD9833_out(uint32_t freq){
	// calculate output word
	/*
	 * freqreg = (fout * 2^28)/fclk
	 */
	uint8_t 	i, SPCR_old, freqOut1, freqOut2, freqOut3, freqOut4;
	uint64_t 	freqWord;
	// Initialize SPI for communication with AD9833
	/* Set SPI Control Register:
	 * SPIE=SPI interrupt enable;	SPE=SPI Enable; 	MSTR=Master(1)/Slave;	DORD= Data Order: LSB first(1)/ MSB first;
	 * CPHA=Clock Phase: Leading/Trailing(1);	CPOL=clock idle polarity;	SPR1=SPI clock rate */
	SPCR_old = SPCR;
	SPCR = (0<<SPIE) | (1<<SPE) | (1<<MSTR) | (0<<DORD) | (0<<CPHA) | (1<<CPOL) | (0<<SPR1) | (0<<SPR0);
	SPDR = 0x00;
	while(!(SPSR & (1<<SPIF)));		// IMPORTANT: Wait for display's transmission to finish
	//SPCR = (0<<SPIE) | (1<<SPE) | (1<<MSTR) | (1<<DORD) | (0<<CPHA) | (1<<CPOL) | (0<<SPR1) | (0<<SPR0);
	PORTB |= (1<<PB2);
	freqWord = freq;
	freqWord = (freqWord * 268435456);
	freqWord = (freqWord /250000000);
	i = ((freqWord >> 8) | 0b01000000);
	freqOut1 = (i &= ~0b10000000);	// 1st byte- 01+ first 6 bits of LSB word
	freqOut2 = (freqWord);			// 2nd byte- next 8 bits of LSB word
	i = ((freqWord >> 22) | 0b01000000);
	freqOut3 = (i &= ~0b10000000);	// 3rd byte- 01+ first 6bits of MSB word
	freqOut4 = (freqWord >>14);		// 4th byte- next 8 bits of MSB word
	PORTB &= ~(1<<PB2);				// Fsync low (word is being loaded)
	//_delay_us(10);
	// two frequency words:
	SPDR = 0b00100000;				// data register 1/2- two freqreg words, reset=0
	//SPDR = 0b00100001;				// data register 1/2- two freqreg words, reset=1
	while(!(SPSR & (1<<SPIF)));			// Wait until SPI interrupt flag=1
	SPDR = 0b00000000;				// data register 2/2
	while(!(SPSR & (1<<SPIF)));
	SPDR = freqOut1;				// Freqreg 1/4 (1/2 LSB first)
	while(!(SPSR & (1<<SPIF)));
	SPDR = freqOut2;				// Freqreg 2/4 (2/2 LSB)
	while(!(SPSR & (1<<SPIF)));
	SPDR = freqOut3;				// Freqreg 3/4 (1/2 MSB)
	while(!(SPSR & (1<<SPIF)));
	SPDR = freqOut4;				// Freqreg 4/4 (2/2 MSB)
	while(!(SPSR & (1<<SPIF)));
	/*
	SPDR = 0xC0;					// Phasereg 1/2
	while(!(SPSR & (1<<SPIF)));
	SPDR = 0x00;					// Phasereg 2/2
	while(!(SPSR & (1<<SPIF)));
	//*/
	SPDR = 0b00100000;				// exit reset 1/2
	while(!(SPSR & (1<<SPIF)));
	SPDR = 0b00000000;				// exit reset 2/2
	while(!(SPSR & (1<<SPIF)));
	_delay_us(10);
	PORTB |= (1<<PB2);				// FSYNC high- inactive
	_delay_us(10);
	SPCR = SPCR_old;
}

// table for getting numbers from 4x4 keypad; 0xF for NaN
uint8_t table_keypad[16] = {
	1,	2,	3,	0xF,
	4,	5,	6,	0xF,
	7,	8,	9,	0xF,
	0xF,0,	0xF,0xF
};

void max7219_out(uint8_t address, uint8_t data){
	uint8_t SPCR_old = SPCR;
	/* Set SPI Control Register:
	 * SPIE=SPI interrupt enable;	SPE=SPI Enable; 	MSTR=Master(1)/Slave;	DORD= Data Order: LSB first(1)/ MSB first;
	 * CPHA=Clock Phase: Leading/Trailing(1);	CPOL=clock idle polarity;	SPR1=SPI clock rate */
	SPCR = (0<<SPIE) | (1<<SPE) | (1<<MSTR) | (0<<DORD) | (0<<CPHA) | (0<<CPOL) | (0<<SPR1) | (0<<SPR0);
	_delay_us(1);
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE){
		//while(!(SPSR & (1<<SPIF)));			// Wait until SPI interrupt flag=1
		SPDR = address;
		while(!(SPSR & (1<<SPIF)));			// Wait until SPI interrupt flag=1
		SPDR = data;
		while(!(SPSR & (1<<SPIF)));			// Wait until SPI interrupt flag=1
		PORTB |= (1 << 0);
		_delay_us(1);
		PORTB &= ~(1 << 0);
		SPCR = SPCR_old;
		//SPDR = 0x00;
	}
}

void max7219_init(){
	int i;
	// 0x00 to 0x0F-test_reg:
	max7219_out(0x0F, 0);
	// 0x0F to 0x0B-scan_reg:
	max7219_out(0x0B, 0x0F);
	// 0x0F to 0x0A-intensity_register:
	max7219_out(0x0A, 0x02);
	// 0x01 to 0x0C-shutdown_register:
	max7219_out(0x0C, 0x01);
	for(i = 0; i < 8; i++){
		max7219_out(i+1, 0);
	}
	//max7219_out(0x0F, 1);		// 0x01 to TEST 0x0F-register
}

void max7219_show(volatile uint8_t digits[8], uint8_t DP_position){
	uint8_t i, segments;
	for(i = 0; i < 8; i++){
		segments = sedmisegment[digits[i]];
		if(i == DP_position){
			segments |= (1<<7);		// add decimal point to right place
		}
		max7219_out(i + 1, segments);
	}
}

#define debounce 40
volatile uint8_t scanned = 0, new_button_press = 0;

// Interrupt to scan keypad upon state change:
ISR(PCINT2_vect){
	if(PIND == 0b00001111){
		return;		// nothing pressed - interrupt probably triggered by rising edge
	}
	scanned = 0;
	uint16_t buttons = 0;
	uint8_t i = 0;
	_delay_ms(debounce);
	if(PIND == 0b00001111){
		return;
	}
	PCMSK2 |= 0b00000000;	// disable PC interrupts in PCI mask2
	DDRD = 0b00010000;		// row 1 as output
	PORTD = 0b00001111;		// rows output 0; columns pullups
	// check rows one by one:
	for(i = 0; i < 4; i++){
		DDRD = _BV(4 + i);
		_delay_ms(1);
		buttons |= ((~((uint16_t) PIND | 0b1111111111110000)) << (4 * i));		// write row
		//DDRD = (DDRD << 1 );	// check next row
	}
	if(buttons == 0){
		return;
	}
	scanned = 1;
	while(scanned <= 16){
		if((buttons & 1)){
			break;
		}
		else {
			buttons = (buttons >> 1);
			scanned++;
		}
	}
	DDRD = 0b11110000;		// Rows to output 0, columns inputs
	PCMSK2 = 0b00001111;	// PinChange Mask 2: Trigger interrupts by columns being pulled low
	PCIFR = _BV(PCIF2);		// Zero-out PCInt flag just in case
	new_button_press = 1;
}

// digits to display and frequency cursor position:
volatile uint8_t disp_freq[8] = {0, 0, 0, 0, 0, 0, 0, 0};
volatile uint8_t cursor_freq_position = 0;

void disp_fix_overflow(){
	uint8_t j, disp_full = 1, i = cursor_freq_position;
	while((i < 8) && (disp_freq[i] == 0xA)){
		for(j = i+1; j < 8; j++){
			if(disp_freq[j] != 9){
				disp_full = 0;
			}
		}
		if(disp_full){
			disp_freq[i] = 9;
			return;
		}
		else {
			disp_freq[i] = 0;
		}
		if(++i < 8){
			disp_freq[i]++;
		}
	}
	while((i < 8) && (disp_freq[i] == (uint8_t)-1)){
		for(j = i+1; j < 8; j++){
			if(disp_freq[j] != 0){
				disp_full = 0;
			}
		}
		if(disp_full){
			disp_freq[i] = 0;
			return;
		}
		else {
			disp_freq[i] = 9;
		}
		if(++i < 8){
			disp_freq[i]--;
		}
	}
}

uint8_t gray_code_2[4] = {0b00, 0b01, 0b11, 0b10};

#define debounce_button 1000	// in timer steps=32us
volatile uint8_t button_enable_1 = 1, button_enable_2 = 1;
volatile uint8_t gray_last, gray_now, last_stable = 0b11, gray_history, new_value, new_cursor_position;
volatile uint16_t TCNT1_last_1 = 0, TCNT1_last_2 = 0;

// Wheel encoder and buttons service routine:
ISR(PCINT1_vect){
	if((PINC & 0b11) != gray_last){	// if gray changed
		gray_now = PINC & 0b11;
		gray_history = (last_stable << 4) | (gray_last << 2) | (gray_now);
		if((gray_history == 0b000111) || (gray_history == 0b111000)){
			// current digit increment:
			last_stable = gray_now & 0b11;
			if(disp_freq[cursor_freq_position] == 9){
				disp_freq[cursor_freq_position]++;
				disp_fix_overflow();
			}
			else {
				disp_freq[cursor_freq_position]++ ;
			}
			new_value = 1;
		}
		if((gray_history == 0b110100) || (gray_history == 0b001011)){
			// current digit decrement:
			last_stable = gray_now & 0b11;
			if(disp_freq[cursor_freq_position] == 0){
				// disp_freq[cursor_freq_position] = 9;
				disp_freq[cursor_freq_position]--;
				disp_fix_overflow();
			}
			else {
				disp_freq[cursor_freq_position]-- ;
			}
			new_value = 1;
		}
		gray_last = gray_now;
	}
	else{
	//*
		if(!(PINC & (1<<PC2))){
			if(button_enable_1 && ((uint16_t)(TCNT1 - TCNT1_last_1) > debounce_button)){
				if(cursor_freq_position > 0){
					cursor_freq_position--;
				}
				else {
					cursor_freq_position = 7;
				}
				new_cursor_position = 1;
				PORTC ^= (1<<PC5);
				TCNT1_last_1 = TCNT1;
				button_enable_1 = 0;
			}
		}
		else{
			if((uint16_t)(TCNT1 - TCNT1_last_1) > debounce_button){
				button_enable_1 = 1;
				TCNT1_last_1 = TCNT1;
			}
		}
		if(!(PINC & (1<<PC3))){
			if(button_enable_2 && ((uint16_t)(TCNT1 - TCNT1_last_2) > debounce_button)){
				if(cursor_freq_position < 7){
					cursor_freq_position++;
				}
				else{
					cursor_freq_position = 0;
				}
				new_cursor_position = 1;
				PORTC ^= (1<<PC5);
				TCNT1_last_2 = TCNT1;
				button_enable_2 = 0;
			}
		}
		else{
			if((uint16_t)(TCNT1 - TCNT1_last_2) > debounce_button){
				button_enable_2 = 1;
				TCNT1_last_2 = TCNT1;
			}
		}
	}
	//*/
}

#define cursor_time_shown 	10000	// step= 32us
#define cursor_time_period	13000

int main(){
	uint32_t freq, now_set = 0;
	uint16_t disp_cursor_time;
	uint8_t disp_cursor_show, disp_cursor_show_last = 0;
	_delay_ms(1);
	DDRB |= (1 << PB1) | (1 << PB2) | (1 << PB0);
	// PB5= clk; PB3= data; PB2= Fsync; PB0= MAX7219-LOAD
	DDRB = (1<<5) | (1<<3) | (1<<2) | (1 << 1) | (1 << 0);
	// PC0-PC3 display anodes; PC5 latch;
	PORTC = (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3);	// wheel inputs | buttons
	DDRD = 0b11110000;		// rows output, columns input
	_delay_us(1);
	gray_last = PINC & 0b11;
	PORTD = 0b00001111;		// rows output 0, pullups on columns
	freq = freq_0;
//	FreqOut(freq);
	TCCR1B 	= (TCCR1B & 0b11111000) | 0b100;	// set 16bit timer to clck/256 = cca 31kHz = cca 32us
	_delay_us(1);
	TCNT1_last_1 = TCNT1;
	TCNT1_last_2 = TCNT1;
	sei();					// global enable interrupts
	PCICR |= (1<<PCIE2) | (1<<PCIE1);	// PinChange Interrupt Control Register |= 1 << PCI enable 2 | 1
	PCMSK1 = 0b00001111;	// PCInterrupt @PORTC enable 0|1=wheel; 2|3=buttons
	PCMSK2 = 0b00001111;	// PinChange Mask 2: Trigger interrupts by columns being pulled low
	// SPI interrupt enable; LSB first; CPOL = 1; Phase = trailing; clk = f/128;
	//SPCR = _BV(SPIE) | _BV(SPE) | _BV(DORD) | _BV(MSTR) | _BV(CPOL) | _BV(CPHA) | _BV(SPR1) | _BV(SPR0);
	//SPDR = 0x00;
	AD9833_out(0);			// initialize AD9833
	PORTB |= (1<<PB1);
	_delay_ms(200);
	PORTB &= ~(1<<PB1);
	max7219_init();
	max7219_show(disp_freq, 1);
	disp_cursor_time = TCNT1;
	while(1){				// verify if physics laws have changed
		disp_cursor_show = 0;
		if((TCNT1 - disp_cursor_time) > cursor_time_shown){
			disp_cursor_show = 1;
		}
		if((TCNT1 - disp_cursor_time) > cursor_time_period){
			disp_cursor_time = TCNT1;
		}
		if(disp_cursor_show && !disp_cursor_show_last){
			max7219_out(cursor_freq_position + 1, sevens_underscore);
			disp_cursor_show_last = 1;
		}
		else if(!disp_cursor_show && disp_cursor_show_last){
			max7219_show(disp_freq, 1);
			disp_cursor_show_last = 0;
		}
		if(new_value || new_button_press/* && (scanned != 0)*/){
			if(new_button_press){
				disp_freq[cursor_freq_position] = table_keypad[scanned-1];
				if(cursor_freq_position == 0){
					cursor_freq_position = sizeof(disp_freq)-1;
				}
				else {
					cursor_freq_position--;
				}
				new_button_press = 0;
			}
			freq = ((uint32_t)disp_freq[7] * 10000000) + ((uint32_t)disp_freq[6] * 1000000) + ((uint32_t)disp_freq[5] * 100000) + \
					((uint32_t)disp_freq[4] * 10000) + ((uint32_t)disp_freq[3] * 1000) + ((uint32_t)disp_freq[2] * 100) +  \
					((uint32_t)disp_freq[1] * 10) + ((uint32_t)disp_freq[0] * 1);
			PORTB ^= (1 << PB1);
			max7219_show(disp_freq, 1);
			AD9833_out(freq);
			disp_cursor_time = TCNT1;
			now_set = freq;
			new_value = 0;
		}
	}
}





